
(****************************************************************************)
(*                                                                          *)
(*   This program is released to the public domain, all I ask is that you   *)
(*   credit me for my initial programming effort.                           *)
(*                                                                          *)
(*   ...If you have any comments that you'd like to pass on to me, please   *)
(*   write me (via "snail mail"):                                           *)
(*                                                                          *)
(*      Guy McLoughlin,                                                     *)
(*      P.O. Box 519,                                                       *)
(*      Station A, Toronto,                                                 *)
(*      Ontario, Canada  M5W 1E4                                            *)
(*                                                                          *)
(*  This source-code requires the following units to compile:               *)
(*     Reuben Sumner's public domain TP "Search" unit.                      *)
(*     Edwin T. Floyd's public domain (see author's notes) TP "Crc" unit.   *)
(*                                                                          *)
(****************************************************************************)


 {$A+,B+,D-,E-,F-,I+,N-,O-,R-,S-,V+}
 {$M 4096, 0, 0}

                   (* Program to encode/decode binary files.                *)
program Xx3401;
uses
  Xx34Data,
  Xx34Rout,
  Dos,
  Search;

                   (* Display the correct syntax for XX34.                  *)
  procedure Syntax;
  begin
    writeln(CrLf, '    XX34  (Binary/XX-format Converter)  Version 0.1');
          writeln('                 by Guy McLoughlin  01-27-92');
    writeln(CrLf, '      (Syntax)');
          writeln('          XX34 [E,D] <filename> <columns> <rows> [AD1..AD9]');
    writeln(CrLf, '               [E,D] - Encode = E   Decode = D');
          writeln('          <filename> - Any valid filename');
    writeln(CrLf, '      (Encode options)');
          writeln('           <columns> - Min 60  Max 100  (Default 68)');
          writeln('              <rows> - Min 10  Max 200  (Default 85)');
          writeln('          [AD1..AD9] - Alternative XxBlock Delimiter');
    writeln(CrLf, '      (Sample Useage)');
          writeln('           XX34 E MYFILE.ZIP        ( Xx-Encode MYFILE.ZIP       )');
          writeln('           XX34 D MYFILE.XX         ( Xx-Decode MYFILE.XX        )');
    writeln(CrLf, '           XX34 E MYFILE.ZIP 72 90 AD1');
          writeln('              ( Xx-Encode MYFILE.ZIP into XxBlocks 72 chars wide )');
          writeln('              ( by 90 Rows long, using alternative delimiter 1.  )');
    halt(0)
  end;

                   (* Close InFile and OutFile properly.                    *)
  procedure CloseFiles;
  begin
                   (* If the InFile is open, then close it.                 *)
    if InFileOpen then
      close(InFile);

                   (* If the InFile was "READ-ONLY", then reset it again.   *)
    if ReadOnlyIn then
      begin
        setfattr(InFile, (Fattr or 1));
        if (doserror <> 0) then
          begin

                   (* Same as "ErrorMessage(3)".                            *)
            writeln(CrLf2, '  ERROR SETTING FILE-ATTRIBUTE');
            halt(3)
          end
      end;

                   (* If the OutFile is open, then close it.                *)
    if OutFileOpen then
      begin

                   (* If we are "decoding", then...                         *)
        if (Encoding = false) then
          begin

                   (* If the OutFile is NOT corrupt, then...                *)
            if (EraseOutFile = false) then
              begin

                   (* Set the file-date to the original binary's date.      *)
                setftime(OutFile, FileTime);
                if (doserror <> 0) then
                  begin
                    writeln(CrLf2, '  ERROR SETTING FILE-DATE/TIME');
                    halt(20)
                  end
                else

                   (* Else NO DOS errors...                                 *)
                  close(OutFile)
              end
            else

                   (* Else the OutFile IS corrupt. Erase it from disk.      *)
              begin
                close(OutFile);
                erase(OutFile);
                writeln('  FILE ERASED ---> ', OutFileName);
                halt(21)
              end
          end
        else

                   (* Else we are "encoding"...                             *)
          close(OutFile)
      end;

                   (* If the OutFile was "READ-ONLY", then reset it again.  *)
    if ReadOnlyOut then
      begin
        setfattr(OutFile, (Fattr or 1));
        if (doserror <> 0) then
          begin

                   (* Same as "ErrorMessage(3)".                            *)
            writeln(CrLf2, '  ERROR SETTING FILE-ATTRIBUTE');
            halt(3)
          end
      end;
    writeln
  end;

                   (* Display an error message and halt the program.        *)
  procedure ErrorMessage(MessNum : byte);
  begin
    writeln(CrLf);

                   (* Display the correct error message.                    *)
    case MessNum of
       1 : writeln('  FILE NOT FOUND ---> ', InFilename);
       2 : writeln('  ZERO BYTE FILE (CONTAINS NO DATA)');
       3 : writeln('  ERROR SETTING FILE-ATTRIBUTE');
       4 : writeln('  ERROR READING FILE');
       5 : writeln('  XX BLOCK-SIZE IS TOO SMALL TO ENCODE ---> ',
                                                                  InFilename);
       6 : writeln('  INSUFFICIENT DISK SPACE TO CREATE NEW XX ENCODED FILE');
       7 : writeln('  FILE CREATION ERROR');
       8 : writeln('   CANNOT OVERWRITE FILE TO ENCODE');
       9 : writeln('  XX34 HEADER NOT FOUND');
      10 : writeln(' INVALID XX34 HEADER FORMAT ---> BLOCK ',
                                                          (TempBlockNum + 1));
      11 : writeln('  XX HEADER FILE SIZES DO NOT MATCH ---> BLOCK ',
                                                            FileSizeFailByte);
      12 : writeln('  XX HEADER FILE DATES DO NOT MATCH ---> BLOCK ',
                                                            FileDateFailByte);
      13 : writeln('  XX HEADER COLUMN/ROW SIZES DO NOT MATCH ---> BLOCK ',
                                                         XxBlockSizeFailByte);
      14 : writeln('  XX HEADER FILE NAMES DO NOT MATCH ---> BLOCK ',
                                                            FileNameFailByte);
      15 : writeln('  INSUFFICIENT DISK SPACE TO DECODE XX-FILE');
      16 : writeln('   CANNOT OVERWRITE FILE TO DECODE');
      17 : writeln('  ERROR CONVERTING DATE-STRING ---> DATE-RECORD');
      18 : writeln('  CRC FAILURE ---> BLOCK ', TempBlockNum);
      19 : writeln('  BLOCK NUMBER GREATER THAN BLOCK TOTAL ---> PHYSICAL POSITION ',
                                                                 XxBlockFound);
   (*NOTE: THE FOLLOWING ERROR CODES ARE HANDLED ELSE WHERE IN THIS PROGRAM *)
   (* 20 : ERROR SETTING FILE-DATE/TIME                                     *)
   (* 21 : OUT-FILE IS CORRUPT, OUT-FILE ERASED                             *)
   (* 22 : ERROR WITH XX-BLOCK TOTALS                                       *)
   (* 23 : DUPLICATE OR MISSING XX-BLOCK                                    *)

      24 : writeln('  FILE IS TOO LARGE TO ENCODE ---> ', InFileName)
    end;

                   (* Now close the InFile and OutFile properly.            *)
    CloseFiles;
    halt(MessNum)
  end;

                   (* Process the command-line parameters.                  *)
  procedure ProcessParams;
  begin
                   (* If there are too many or too few parameters, then...  *)
    if (paramcount > 5)
    or (paramcount < 2) then

                   (* Display the correct program syntax.                   *)
      Syntax
    else
      begin
                   (* Assign the name of the file to be processed to the    *)
                   (* InFilename variable, in uppercase.                    *)
        InFilename := UpStr(paramstr(2));

                   (* Delete any drive/directory characters from the name.  *)
        while (pos(':', InFilename) <> 0)
          or (pos('\', InFilename) <> 0) do
            delete(InFilename, 1, 1);

                   (* Check to see if we are "encoding" the InFile.         *)
        if (UpStr(paramstr(1)) = 'E') then
          begin
                   (* Yes, we are encoding.                                 *)
            Encoding := true;

                   (* Create the OutFile name, from the InFile's name.      *)
            if (pos('.', InFilename) <> 0) then
              OutFilename := (copy(InFilename, 1, pos('.', InFilename)) + 'XX')
            else
              OutFilename := (copy(InFilename, 1, 8) + '.XX');

                   (* If a valid alternative-delimiter is needed, then...   *)
            if (paramcount > 4) then

                   (* Determine and set the alternative delimiter chosen.   *)
              case pos(UpStr(paramstr(5)), AdString) of
                 1 : begin
                       XxBlockID[1] := AdSet[1];
                       XxHeader1[1] := AdSet[1];
                       XxEndMarker := ConvertChar(XxEndMarker, #42, AdSet[1])
                     end;
                 4 : begin
                       XxBlockID[1] := AdSet[2];
                       XxHeader1[1] := AdSet[2];
                       XxEndMarker := ConvertChar(XxEndMarker, #42, AdSet[2])
                     end;
                 7 : begin
                       XxBlockID[1] := AdSet[3];
                       XxHeader1[1] := AdSet[3];
                       XxEndMarker := ConvertChar(XxEndMarker, #42, AdSet[3])
                     end;
                10 : begin
                       XxBlockID[1] := AdSet[4];
                       XxHeader1[1] := AdSet[4];
                       XxEndMarker := ConvertChar(XxEndMarker, #42, AdSet[4])
                     end;
                13 : begin
                       XxBlockID[1] := AdSet[5];
                       XxHeader1[1] := AdSet[5];
                       XxEndMarker := ConvertChar(XxEndMarker, #42, AdSet[5])
                     end;
                16 : begin
                       XxBlockID[1] := AdSet[6];
                       XxHeader1[1] := AdSet[6];
                       XxEndMarker := ConvertChar(XxEndMarker, #42, AdSet[6])
                     end;
                19 : begin
                       XxBlockID[1] := AdSet[7];
                       XxHeader1[1] := AdSet[7];
                       XxEndMarker := ConvertChar(XxEndMarker, #42, AdSet[7])
                     end;
                22 : begin
                       XxBlockID[1] := AdSet[8];
                       XxHeader1[1] := AdSet[8];
                       XxEndMarker := ConvertChar(XxEndMarker, #42, AdSet[8])
                     end;
                25 : begin
                       XxBlockID[1] := AdSet[9];
                       XxHeader1[1] := AdSet[9];
                       XxEndMarker := ConvertChar(XxEndMarker, #42, AdSet[9])
                     end
              else
                   (* Else parameter for the alternative-delimiter was not  *)
                   (* correct. Display the program's syntax.                *)
                Syntax
              end; (* Of "case pos(UpStr(paramstr(5)), AdString)".          *)

                   (* If a different XxBlock row size is required then...   *)
            if (paramcount > 3) then
              begin
                   (* Use the "RowColTable" to determine the row size.      *)
                TempRowSize := (StrPos(RowColTable, sizeof(RowColTable),
                                                      paramstr(4)) + 3) div 4;

                   (* If the XxBlock row-size is too small or too big, then *)
                   (* use the default size.                                 *)
                If (TempRowSize < MinRow)
                  or (TempRowSize > MaxRow) then
                    TempRowSize := RowDefault;

                   (* Assign the new data to the initialized XxHeader.      *)
                XxHeader2.FRows := RowColTable[TempRowSize].Num
              end;

                   (* If a different XxBlock column size is required then...*)
            if (paramcount > 2) then
              begin
                   (* Use the "RowColTable" to determine the column size.   *)
                TempColSize := (StrPos(NumTable, sizeof(NumTable),
                                                      paramstr(3)) - 1) div 3;

                   (* Make sure the column-width is a multiple of 4.        *)
                TempColSize := (TempColSize div 4) * 4;

                   (* If the XxBlock column-size is too small or too big,   *)
                   (* then use the default size.                            *)
                If (TempColSize < MinCol)
                  or (TempColSize > MaxCol) then
                    TempColSize := ColDefault;

                   (* Assign the new data to the initialized XxHeader.      *)
                XxHeader2.FCols := RowColTable[TempColSize].Num
              end
          end
        else
                   (* Else if the first parameter does not indicate         *)
                   (* "decoding", then display program syntax.              *)
          if (UpStr(paramstr(1)) <> 'D') then
            Syntax
      end
  end;

                   (* Write the next XxHeader to the XxBuffer.              *)
  procedure WriteHeader;
  var
    TempStr : Str12;
  begin
                   (* Clear the temp-string.                                *)
    fillchar(TempStr, sizeof(TempStr), 0);

                   (* If this is the last BinBlock, then determine          *)
                   (* actual binary-block size. Instead of using the        *)
                   (* size calculated from the column/row sizes.            *)
    if ((BinFileIndex + BinBuffIndex + BinBlockSize) > BinFilesize) then
      begin
        BinBlockSize := (BinFilesize - (BinFileIndex + BinBuffIndex));
        if (BinBlockSize mod 3) <> 0 then
          BinBlockSize := ((BinBlockSize div 3) + 1) * 3;
      end;
                   (* Calculate the 16-bit Crc value for the binary block,  *)
                   (* convert this to a string, and convert all ' ' to '0'. *)
    TempStr := ConvertChar(Num2Str(CalcCrc, 5), #32, #48);

                   (* Write this XxHeader to the XxBuffer.                  *)
    move(TempStr[1], XxHeader2.CrcValue, 5);

                   (* Advance the block number by 1.                        *)
    inc(TempBlockNum);

                   (* Set the XxHeader's block number.                      *)
    XxHeader2.Block := RowColTable[TempBlockNum].Num;

                   (* Assign this data to the initialized XxHeader.         *)
    move(XxHeader2, XxBuffer[XxBuffIndex + 1], sizeof(XxHeader2));

                   (* Diplay which block we are encoding.                   *)
    write(TempBlockNum:4);

                   (* Advance the XxBuffer indexes.                         *)
    inc(XxBuffIndex, sizeof(XxHeader2));
    inc(NonXxIndex, sizeof(XxHeader2));
    LineIndex := 1
  end;


  procedure WriteFooter;
  begin
                   (* Write the next XxFooter to the XxBuffer.              *)
    move(XxEndMarker[1], XxBuffer[XxBuffIndex + 1], length(XxEndMarker));

                   (* Advance the XxBuffer indexes.                         *)
    inc(XxBuffIndex, length(XxEndMarker));
    inc(NonXxIndex, length(XxEndMarker))
  end;

                   (* Load up the XxBuffer with XxBlocks.                   *)
  procedure FillXxBuffer;
  begin
                   (* Clear Xx-buffer.                                      *)
    fillchar(XxBuffer, sizeof(XxBuffer), 0);

                   (* Position file-pointer for next Block to read in.      *)
    if (XxFileIndex <> 0) then
      seek(InFile, XxFileIndex);

                   (* Determine the number of bytes to read into XxBuffer.  *)
    if (XxBuffSize > (XxFilesize - XxFileIndex)) then
      BytesToGet := (XxFilesize - XxFileIndex)
    else
      BytesToGet := XxBuffSize;

                   (* Read from Xx-file.                                    *)
    blockread(InFile, XxBuffer, BytesToGet);

                   (* Re-intialize XxBuffer variables.                      *)
    BytesInXxBuff := BytesToGet;
    XxBuffIndex := 0;
    XxBuffOffset := 0
  end;

                   (* Open the files to process and create.                 *)
  procedure OpenFiles;
  begin
    assign(InFile, InFilename);

                   (* Get the InFile's file-attributes.                     *)
    getfattr(InFile, Fattr);
    if (doserror <> 0) then
      ErrorMessage(1);

                   (* If InFile is "READ-ONLY", record this and turn it off.*)
    if ((Fattr and 1) <> 0) then
      begin
        ReadOnlyIn := true;
        setfattr(InFile, (Fattr xor 1));
        if (doserror <> 0) then
          ErrorMessage(3)
      end;
                   (* Try to open the InFile.                               *)
    {$I-}
    reset(InFile, 1);
    {$I+}
    if (ioresult <> 0) then
      ErrorMessage(4)
    else
                   (* Indicate that InFile is now open.                     *)
      InFileOpen := true;

                   (* Check if this is a "zero-byte" file.                  *)
    if (filesize(InFile) = 0) then
      ErrorMessage(2);

                   (* If we are "encoding" then...                          *)
    if Encoding then
      begin
                   (* Check to see if the file to be encoded is too big.    *)
        if (filesize(InFile) > MaxFileSize) then
          ErrorMessage(24);

                   (* Assign XxBlockSize.                                   *)
        XxBlockSize := TempColSize * (TempRowSize - 1);

                   (* BinBlockSize is 3/4 of XxBlockSize.                   *)
        BinBlockSize := (XxBlockSize div 4) * 3;

                   (* Assign the binary-filesize.                           *)
        BinFilesize := filesize(InFile);

                   (* Check if the binary-file is too big.                  *)
        if (BinFileSize > (longint(BinBlockSize) * BlockMax)) then
          ErrorMessage(5);

                   (* Obtain the file-date information.                     *)
        getftime(InFile, FileTime);
        unpacktime(FileTime, FileDate);

                   (* Convert file-date info to string format.              *)
        with FileDate do
          begin
            TempFileDate[1] := NumTable[(day + 1)].Num;
            TempFileDate[2] := NumTable[(month + 1)].Num;
            if (year < NextCentury) then
              TempFileDate[3] := NumTable[((year  - ThisCentury) + 1)].Num
            else
              TempFileDate[3] := NumTable[((year  - NextCentury) + 1)].Num
          end;

                   (* Assign filedate to the initialized XxHeader.          *)
        move(TempFileDate[1], XxHeader2.Fday, 6);

               (* Convert filesize to string format, and convert ' ' to '0' *)
        TempStr1 := ConvertChar(Num2Str(BinFilesize, 6), #32, #48);

                   (* Assign filesize to the initialized XxHeader.          *)
        move(TempStr1[1], XxHeader2.Size, 6);

                   (* Assign filename to initialized XxHeader.              *)
        move(InFileName[1], XxHeader2.Filename[1 + (12 - length(InFileName))],
                                                          length(InFileName));

                   (* Determine the total number of BinaryBlocks.           *)
        if ((BinFilesize mod BinBlockSize) <> 0) then
          TempBlockTotal := ((BinFilesize div BinBlockSize) + 1)
        else
          TempBlockTotal := (BinFilesize div BinBlockSize);

                   (* Check to see if there is disk-space for Xx-file.      *)

                   (* Additional 102 bytes assumed for each Xx-block.       *)
                   (* Account for size of CrLf's for each Xx-block row.     *)
                   (* Leave last cluster free. (HD large cluster size used) *)
        if (BinFilesize mod 3 <> 0) then
          TempIndex1 := ((((BinFilesize div 3) + 1) * 4) +
                         (TempBlockTotal * XxOverHead) +
                         ((((BinFilesize div 3) + 1) * 8) div TempColSize) +
                         LargeCluster)
        else
          TempIndex1 := (((BinFilesize div 3) * 4) +
                         (TempBlockTotal * XxOverHead) +
                         (((BinFilesize div 3) * 8) div TempColSize) +
                         LargeCluster);
        if (TempIndex1 > diskfree(0)) then
          ErrorMessage(6);

                   (* Assign the total number of BinaryBlocks to XxHeader.  *)
        XxHeader2.BlockTot := RowColTable[TempBlockTotal].Num;

                   (* Open/create the new Xx-file.                          *)
        assign(OutFile, OutFilename);

                   (* Get the OutFile's file-attributes.                    *)
        getfattr(OutFile, Fattr);
        if (doserror = 0) then

                   (* If OutFile file-attribute is 'READ-ONLY, then...      *)
          if ((Fattr and 1) <> 0) then
            begin
                   (* Record OutFile's 'READ-ONLY' state.                   *)
              ReadOnlyOut := true;

                   (* Turn OutFile's 'READ-ONLY' attribute off.             *)
              setfattr(OutFile, (Fattr xor 1));
              if (doserror <> 0) then
                ErrorMessage(3)
            end;
                   (* Try to open the OutFile, and indicate your success    *)
                   (* with the "OutFileOpen" boolean.                       *)
        {$I-}
        reset(OutFile, 1);
        {$I+}
                   (* If OutFile doesn't exist then...                      *)
        if (ioresult <> 0) then
          begin
                   (* Try to create OutFile.                                *)
            {$I-}
            rewrite(OutFile, 1);
            {$I+}
            if (ioresult <> 0) then
              ErrorMessage(7)
            else
              OutFileOpen := true
          end
        else       (* Else, the file does exist.                            *)
          begin
            OutFileOpen := true;

                   (* Check to see if it's OK to over-write this file.      *)
            write(CrLf, ' Overwrite ', OutFilename, ' file ??? (Y/N) ');
            if YesNo then
                   (* Yes it's OK to overwrite this file.                   *)
              begin
                   (* You cannot overwrite the file you want to encode.     *)
                if (InFilename = OutFilename) then
                  ErrorMessage(8);
                writeln;
                {$I-}
                rewrite(OutFile, 1);
                {$I+}
                if (ioresult <> 0) then
                  ErrorMessage(7)
              end
            else   (* Else, NO it's NOT OK to overwrite this file. ABORT!   *)
              begin
                writeln(CrLf, CrLf, ' Encoding Aborted');
                CloseFiles;
                halt(0)
              end
          end
      end
    else           (* Decoding = true.                                      *)

                   (* Assign size of InFile.                                *)
      XxFilesize := filesize(InFile)
  end;

                   (* Encode complete binary file using XX34 format.        *)
  procedure Encode34;
  begin
                   (* Display encoding message.                             *)
    writeln(CrLf, ' ENCODING BLOCK :');

                   (* Write 2 CrLf's to the beginning of the new Xx-file.   *)
    blockwrite(OutFile, CrLf2, sizeof(CrLf2));

    repeat         (* Until the complete binary-file has been processed.    *)

                   (* Clear Binary-buffer and Xx-buffer.                    *)
      fillchar(BinBuffer, sizeof(BinBuffer), 0);
      fillchar(XxBuffer, sizeof(XxBuffer), 0);

                   (* Determine the number of binary-bytes to read in.      *)
                   (* If the complete binary can be read in, then...        *)
      if (BinBuffSize > (BinFilesize - BinFileIndex)) then
        BytesToGet := (BinFilesize - BinFileIndex)
      else
                   (* Else ONLY read-in a multiple of BinBlockSize bytes.   *)
        BytesToGet := (BinBuffSize div BinBlockSize) * BinBlockSize;

                   (* Read from the Binary-file.                            *)
      blockread(InFile, BinBuffer, BytesToGet);

                   (* Initialize the buffer variables.                      *)
      BinBuffIndex := 0;
      XxBuffIndex := 0;
      NonXxIndex := 0;

                   (* Write the Xx-block header.                            *)
      WriteHeader;

      repeat       (* ...Until all bytes in the Binary-buffer are processed.*)

                   (* Encode 3 binary-bytes and write 4 Xx-chars.           *)
        XxBuffer[XxBuffIndex + 1] :=
                                  Xxset[(BinBuffer[BinBuffIndex + 1] shr 2)];
        XxBuffer[XxBuffIndex + 2] :=
                            Xxset[((BinBuffer[BinBuffIndex + 1] and 3) shl 4)
                                      + (BinBuffer[BinBuffIndex + 2] shr 4)];
        XxBuffer[XxBuffIndex + 3] :=
                           Xxset[((BinBuffer[BinBuffIndex + 2] and 15) shl 2)
                                      + (BinBuffer[BinBuffIndex + 3] shr 6)];
        XxBuffer[XxBuffIndex + 4] :=
                                 Xxset[(BinBuffer[BinBuffIndex + 3] and 63)];

                   (* Advance the Binary-index and Xx-Index.                *)
      inc(BinBuffIndex, 3);
      inc(XxBuffIndex, 4);

                   (* if you've completed an Xx-row, then...                *)
        if ((XxBuffIndex - NonXxIndex) mod TempColSize = 0) then
          begin
                   (* Add a CrLf to the Xx-buffer.                          *)
            XxBuffer[XxBuffIndex + 1] := #13;
            XxBuffer[XxBuffIndex + 2] := #10;

                   (* Advance the Xx-index, Non-Xx-index, and the LineIndex.*)
            inc(XxBuffIndex, 2);
            inc(NonXxIndex, 2);
            inc(LineIndex, 1);

                   (* if Xx-block is completely processed then...           *)
            if (LineIndex = TempRowSize) then
              begin
                   (* Write the XxFooter to Xx-buffer.                      *)
                WriteFooter;

                   (* Reset LineIndex.                                      *)
                LineIndex := 0;

                   (* If another Binary-block is waiting to be processed... *)
                if (BinBuffIndex < BytesToGet) then
                  WriteHeader
              end
          end;

                   (* Until all bytes in the Binary-buffer are processed.   *)
      until (BinBuffIndex >= BytesToGet);

                   (* Advance the binary-file index.                        *)
      inc(BinFileIndex, BytesToGet);

                   (* Write out the Xx-buffer to the OutFile.               *)
      blockwrite(OutFile, XxBuffer[1], XxBuffIndex);

                   (* Until the complete binary-file has been processed.    *)
    until (BinFileIndex >= BinFilesize);

                   (* Add Footer to the end of the new Xx-file, if needed.  *)
    if (LineIndex <> 0) then
      begin
                   (* If the last line was not a complete row, then...      *)
        if ((XxBuffIndex - NonXxIndex) mod TempColSize <> 0) then

                   (* Add a CrLf to the file being written.                 *)
          blockwrite(OutFile, CrLf, sizeof(CrLf));
        blockwrite(OutFile, XxEndMarker[1], length(XxEndMarker))
      end
  end;

                   (* Completely decode the XX-encoded file.                *)
  procedure Decode43;
  var
    Error : integer;
  begin
                   (* Display scanning message.                             *)
    write(CrLf, ' SCANNING FOR BLOCK IDENTIFIER');

                   (* Load XxBuffer from XxEncoded file.                    *)
    FillXxBuffer;

                  (* Determine what the XxBlockID is, alternative-delimiter *)
                  (* may have been used. Search only the first 20K of the   *)
                  (* Infile.                                                *)
    TempIndex1 := 1;

    repeat        (* Until a valid XxBlockID is found, or all ten headers   *)
                  (* have been searched for.                                *)

                  (* Search for a valid XxBlockID.                          *)
      HeaderPos := StrPos(XxBuffer, BytesInXxBuff, XxBlockID);
      if (HeaderPos = 0) then
        begin
          XxBlockID[1] := AdSet[TempIndex1];
          inc(TempIndex1, 1)
        end

                  (* Until a valid XxBlockID is found, or all ten headers   *)
                  (* have been searched for.                                *)
    until (TempIndex1 = 11) or (HeaderPos <> 0);

                  (* if NO XxBlockID is found in the first 20K of the       *)
                  (* InFile, halt with error-message.                       *)
    if (HeaderPos = 0) then
      ErrorMessage(9)
    else
                  (* Else XxBlockID is found, reset HeaderPos variable.     *)
      HeaderPos := 0;

                   (* Display new message.                                  *)
    write('   OK', CrLf, ' SCANNING FOR BLOCK HEADERS');

    repeat         (* Repeat until all the Xx-headers have been processed.  *)

                   (* Determine Xx-header position in the XxBuffer.         *)
      HeaderPos := StrPos(XxBuffer[XxBuffOffset + 1],
                      (BytesInXxBuff - (XxBuffOffset + XxHeaderSize)),
                       XxBlockID);

                   (* If another Xx-header is present XxBuffer.             *)
      if (HeaderPos <> 0) then
                   (* Process another Xx-header.                            *)
        begin
                   (* Advance XxBuffOffset and XxBlockFound.                *)
          inc(XxBuffOffset, HeaderPos);
          inc(XxBlockFound, 1);

                   (* Clear XxHeader variable of old data.                  *)
          fillchar(XxHeader1, sizeof(XxHeader1), 0);

                   (* Copy Xx-header found from the XxBuffer to the         *)
                   (* XxHeader1 variable for processing.                    *)
          move(XxBuffer[XxBuffOffset], XxHeader1, sizeof(XxHeader1));

                   (* Assign Xx-header data to appropriate variables.       *)
          with XxHeader2 do
            begin
              TempBinFileSize := Str2Num(Size, 6);
              if (TempBinFileSize = BadNumber) then
                ErrorMessage(10);
              TempFileDate[1] := Fday;
              TempFileDate[2] := Fmonth;
              TempFileDate[3] := Fyear;

                   (* Use "RowColTable" to assign column size from the     *)
                   (* XxHeader data.                                        *)
              TempColSize := (StrPos(RowColTable, sizeof(RowColTable),
                              FCols) + 3) div 4;

                   (* Use "RowColTable" to assign row size from the         *)
                   (* XxHeader data.                                        *)
              TempRowSize  := (StrPos(RowColTable, sizeof(RowColTable),
                              Frows) + 3) div 4;
              TempFilename := FileName;

                   (* Use "RowColTable" to assign XxBlock number from the   *)
                   (* XxHeader data.                                        *)
              TempBlockNum := (StrPos(RowColTable, sizeof(RowColTable), Block)
                               + 3) div 4;

                   (* Use "RowColTable" to assign XxBlock total from        *)
                   (* XxHeader data.                                        *)
              TempBlockTotal := (StrPos(RowColTable, sizeof(RowColTable),
                                        BlockTot) + 3) div 4;

                   (* Record the XxHeader's Crc value from the XxHeader     *)
                   (* data.                                                 *)
              CrcRec[TempBlockNum] := Str2Num(CrcValue, 5)
            end;

                   (* Check to see if Xxblock number from the XxHeader is   *)
                   (* valid.                                                *)
          if (TempBlockNum > TempBlockTotal) then
            ErrorMessage(19);

                   (* Record Xx-block number.                               *)
          inc(BlockNumRec[TempBlockNum], 1);

                   (* Record the Xxblock total.                             *)
          inc(BlockTotRec[TempBlockTotal], 1);

                   (* Keep track of largest XxHeader BlockTotal number.     *)
          if (TempBlockTotal > MaxBlockTotal) then
            MaxBlockTotal := TempBlockTotal;

                   (* Assign the XxBlock file position.                     *)
          if (XxFileIndex <> 0) then
            XxBlockPos[TempBlockNum] := (XxBuffOffset + XxFileIndex) - 1
          else
            XxBlockPos[TempBlockNum] := XxBuffOffset - 1;

                   (* Assign the physical (not logical) XxBlock order.      *)
          PBlockOrder[XxBlockFound] := TempBlockNum;

                   (* Calculate Xx-block size.                              *)
          TempBlockSize := TempColSize * (TempRowSize - 1);

                   (* Assign Xx-block size to XxBlockSize array.            *)
        XxBlockSizeRec[TempBlockNum] := TempBlockSize;

                   (* Set and check the XxHeader error-checking variables.  *)

                   (* If this is physically the very first XxBlock then...  *)
                   (* Set the temporary XxBlock variables.                  *)
          if (XxBlockFound = 1 ) then
            begin
              BinFilesize := TempBinFilesize;
              BinFileDate := TempFileDate;
              XxBlockSize := TempBlockSize;
              OutFilename := TempFilename
            end
          else
                   (* Else this is NOT the first XxBlock then check for     *)
                   (* XxHeader errors. Set the XxBlock number if an error   *)
                   (* did occur.                                            *)
            begin
              if (TempBinFileSize <> BinFilesize)
              and (FileSizeFail = false) then
                FileSizeFailByte := TempBlockNum;

              if ((TempFileDate[1] <> BinFileDate[1])
              or (TempFileDate[2] <> BinFileDate[2])
              or (TempFileDate[3] <> BinFileDate[3]))
              and (FileDateFail = false) then
                FileDateFailByte := TempBlockNum;

              if (TempBlockSize <> XxBlockSize)
              and (XxBlockSizeFail = false) then
                XxBlockSizeFailByte := TempBlockNum;

              If (TempFilename <> OutFilename)
              and (FileNameFail = false) then
                FileNameFailByte := TempBlockNum;

              if (BlockNumRec[TempBlockNum] > 1) then
                inc(BlockNumRec[0], 1)
            end;

                   (* Advance the XxBuffer index.                           *)
          if (XxBuffOffset < (BytesInXxBuff - XxHeaderSize)) then
            inc(XxBuffOffset, XxHeaderSize)
        end
      else         (* HeaderPos = 0 , No addittional Xx-headers found.      *)

                   (* Advance the InFile index, and refill the XxBuffer.    *)
        if ((XxFileIndex + XxBuffOffset + TempBlockSize) < XxFilesize) then
          begin
            inc(XxFileIndex, (XxBuffOffset + XxHeaderSize));
            FillXxBuffer
          end;

                   (* Until the entire InFile has been scanned.             *)
    until ((XxFileIndex + XxBuffOffset + TempBlockSize) >= XxFilesize);

                   (* Display new message.                                  *)
    write('      OK', CrLf, ' CHECKING FOR BLOCK ERRORS');

                   (* Set error-flag if Xx-block totals do not match.       *)
    if (BlockTotRec[MaxBlockTotal] <> MaxBlockTotal) then
      inc(BlockTotRec[0], 1);

                   (* Check for Xx-blocks errors.                           *)

                   (* Check if the number of XxBlocks found, match XxBlock  *)
                   (* total from XxHeader data.                             *)
    if (BlockTotRec[0] <> 0) then
      begin
        writeln(CrLf);
        if (BlockTotRec[0] <> 0) then
          begin
            writeln(CrLf, ' ERROR WITH XX-BLOCK TOTALS:', CrLf);
            writeln('               BLOCKS FOUND = ', XxBlockFound);
            for TempIndex1 := 1 to MaxBlockTotal do
              if (BlockTotRec[TempIndex1] > 0) then
                writeln(' BLOCK TOTAL FROM XX-HEADER = ', TempIndex1:2)
          end;
        ErrorMessage(22)
      end;

                   (* Check for duplicate/missing XxBlocks.                 *)
    if (BlockNumRec[0] <> 0) then
      begin
        for TempIndex1 := 1 to MaxBlockTotal do
          if (BlockNumRec[TempIndex1] > 1) then
            writeln(CrLf, CrLf, '   DUPLICATE COPIES OF BLOCK ', TempIndex1:2)
          else
            if (BlockNumRec[TempIndex1] = 0) then
              write(CrLf, '               MISSING BLOCK ', TempIndex1:2);
        ErrorMessage(23)
      end;

                   (* If there ARE XxHeader errors, then halt the program.  *)

    if FileSizeFail then
      ErrorMessage(11);

    if FileDateFail then
      ErrorMessage(12);

    if XxBlockSizeFail then
      ErrorMessage(13);

    if FileNameFail then
      ErrorMessage(14);

                   (* Display message.                                      *)
    writeln('       OK');

                   (* Make sure their is room to write the decoded binary-  *)
                   (* file to disk.                                         *)
    if (BinFilesize mod OneKay <> 0) then
      TempIndex1 := (BinFilesize div OneKay) + 1
    else
      TempIndex1 := BinFilesize div OneKay;
    if (TempIndex1 > (diskfree(0) div OneKay)) then
      ErrorMessage(15);

                   (* Remove extra '-' chars from OutFilename.              *)
    while (OutFilename[1] = '-') do
      delete(OutFilename, 1, 1);

                   (* Assign the OutFilename to the OutFile variable.       *)
    assign(OutFile, OutFilename);

                   (* Check to see if the OutFile is 'READ-ONLY'.           *)
    getfattr(OutFile, Fattr);
    if (doserror = 0) then

                   (* If OutFile file-attribute is 'READ-ONLY'.             *)
      if ((Fattr and 1) <> 0) then
        begin
                   (* Record OutFile's 'READ-ONLY' state.                   *)
          ReadOnlyOut := true;

                   (* Turn OutFile's 'READ-ONLY' attribute off.             *)
          setfattr(OutFile, (Fattr xor 1));
          if (doserror <> 0) then
            ErrorMessage(3)
        end;
    {$I-}
    reset(OutFile, 1);
    {$I+}
                   (* If OutFile doesn't exist then...                      *)
    if (ioresult <> 0) then
      begin
                   (* Try to create OutFile.                                *)
        {$I-}
        rewrite(OutFile, 1);
        {$I+}
        if (ioresult <> 0) then
          ErrorMessage(7)
        else
          OutFileOpen := true
      end
    else           (* Else, the file does exist.                            *)
      begin
        OutFileOpen := true;

                   (* Check to see if it's OK to over-write this file.      *)
        write(CrLf, ' Overwrite ', OutFilename, ' file ??? (Y/N) ');
        if YesNo then
                   (* Yes it's OK to overwrite this file.                   *)
          begin
                   (* You cannot overwrite the file you want to decode.     *)
            if (InFilename = OutFilename) then
              ErrorMessage(16);
            writeln;
            {$I-}
            rewrite(OutFile, 1);
            {$I+}
            if (ioresult <> 0) then
              ErrorMessage(7)
          end
        else       (* Else, NO it's NOT OK to overwrite this file. ABORT!   *)
          begin
            writeln(CrLf, CrLf, ' Decoding Aborted');
            ErrorMessage(0)
          end
      end;
                   (* Determine original binary-file date from the XxHeader *)
      with FileDate do
        begin
          val(BinFileDate[3], year, Error);
          if (Error <> 0) then
            ErrorMessage(17);
          inc(year, ThisCentury);
          val(BinFileDate[2], month, Error);
          if (Error <> 0) then
            ErrorMessage(17);
          val(BinFileDate[1], day, Error);
          if (Error <> 0) then
            ErrorMessage(17);
          hour := 0;
          min := 0;
          sec := 0
        end;
      packtime(FileDate, FileTime);

                   (* Determine the maximum XxBlock sizes, and the actual   *)
                   (* size of the last XxBlock.                             *)

                   (* Determine each XxBlock's maximum block-size.          *)
    for TempIndex1 := 1 to XxBlockFound do

                   (* If this XxBlock is not the physically last block...   *)
      if (TempIndex1 <> XxBlockFound) then
        PBlockSize[(PBlockOrder[TempIndex1])] :=
                                  XxBlockPos[(PBlockOrder[TempIndex1 + 1])] -
                                        XxBlockPos[(PBlockOrder[TempIndex1])]

                   (* else, this IS the physically last Xxblock.            *)
      else
                   (* If the last XxBlock is smaller than the buffer-size   *)
        if ((XxFilesize - XxBlockPos[(PBlockOrder[TempIndex1])]) <
            XxBuffSize) then
          PBlockSize[(PBlockOrder[TempIndex1])] := (XxFilesize -
                                       XxBlockPos[(PBlockOrder[TempIndex1])])
                (* Else the last block is NOT smaller than the buffer-size. *)
        else
          PBlockSize[(PBlockOrder[TempIndex1])] := XxBuffSize;

                   (* Calculate the size of the last XxBlock.               *)
    TempIndex2 := 0;
                   (* Add the XxHeader size for all but the last XxBlock.   *)
    for TempIndex1 := 1 to (XxBlockFound - 1) do
      inc(TempIndex2, XxBlockSizeRec[TempIndex1]);

                   (* Convert to binaray-block size.                        *)
    TempIndex1 := BinFileSize - ((TempIndex2 div 4) * 3);

                   (* Set the XxHeader size of the last XxBlock.            *)
    if ((TempIndex1 mod 3) <> 0) then
      XxBlockSizeRec[XxBlockFound] := ((TempIndex1 div 3) + 1) * 4
    else
      XxBlockSizeRec[XxBlockFound] := (TempIndex1 div 3) * 4;

                   (* Time to decode Xx-block file!                         *)

                   (* Display new message.                                  *)
    writeln(CrLf, ' DECODING BLOCK :');

                   (* Initialize variables.                                 *)
    TempIndex1 := 1;
    TempIndex2 := 0;
    TempBlockNum  := 1;
    BinFileIndex := 0;

    repeat         (* Until the binary file has been completely re-built.   *)

                   (* Initialize the buffer variables.                      *)
      BinBuffIndex    := 0;
      BinBytesToWrite := 0;
      XxBlockIndex    := 0;
      XxBuffOffset    := 0;

                   (* Clear binary-buffer.                                  *)
      fillchar(BinBuffer, sizeof(BinBuffer), 0);

                   (* Clear Xx-buffer.                                      *)
      fillchar(XxBuffer, sizeof(XxBuffer), 0);

                   (* Clear the Buffer XxBlock position array.              *)
      fillchar(BuffBlockPos, sizeof(BuffBlockPos), 0);

                   (* Load up the XxBuffer with XxBlocks.                   *)

                   (* While new blocks will fit in the XxBuffer...          *)
      while ((XxBuffOffset + PBlockSize[TempIndex1 + 1]) < XxBuffSize)

                   (* and the last XxBlock has not been read in...          *)
      and (TempIndex2 < XxBlockFound) do
        begin

                  (* Advance the XxBlock number to read into the XxBuffer.  *)
          inc(TempIndex2);

                  (* Determine the number of binary bytes to write (part 1) *)
          inc(BinBytesToWrite, XxBlockSizeRec[TempIndex2]);

                   (* Determine the number of XxFile bytes to read-in.      *)
          BytesToGet := PBlockSize[TempIndex2];

                   (* Position InFile pointer.                              *)
          seek(InFile, XxBlockPos[TempIndex2]);

                   (* Read an XxBlock into the XxBuffer.                    *)
          blockread(InFile, XxBuffer[XxBuffOffset + 1], BytesToGet);

                   (* Record the XxBlock's position in the XxBuffer. Allow  *)
                   (* for the XxHeader size.                                *)
          BuffBlockPos[TempIndex2] := (XxBuffOffset + 61);

                   (* Advance the buffer-offset by the number of bytes read.*)
          inc(XxBuffOffset, BytesToGet);
      end;
                  (* Determine the number of binary bytes to write (part 2) *)
      BinBytesToWrite := (BinBytesToWrite div 4) * 3;

                   (* Set the XxBuffOffset for this XxBlock.                *)
      XxBuffOffset := BuffBlockPos[TempBlockNum];

      repeat       (* Until all XxBlocks in the XxBuffer are processed.     *)

                   (* Display which XxBlock is being decoded.               *)
        write(TempBlockNum:4);

                   (* Initialize XxBlock variables.                         *)
        XxBlockIndex := 0;
        BinBlockSize := 0;

        repeat     (* Until this XxBlock is completely decoded.             *)

                   (* Search for next char in the XxSet.                    *)
          while (FindByte(XxSet, 64, byte(XxBuffer[XxBuffOffset])) = 0) do
            inc(XxBuffOffset, 1);

                   (* Decode 4 Xx-Chars into 3 binary bytes.                *)
          BinBuffer[BinBuffIndex + 1] :=
                                   (Binset[ord(XxBuffer[XxBuffOffset])] shl 2)
                            + (Binset[ord(XxBuffer[XxBuffOffset + 1])] shr 4);
          BinBuffer[BinBuffIndex + 2] :=
                      ((Binset[ord(XxBuffer[XxBuffOffset + 1])] and 15) shl 4)
                            + (Binset[ord(XxBuffer[XxBuffOffset + 2])] shr 2);
          BinBuffer[BinBuffIndex + 3] :=
                       ((Binset[ord(XxBuffer[XxBuffOffset + 2])] and 3) shl 6)
                   + (Binset[ord(XxBuffer[XxBuffOffset + 3])] and 63);

                   (* Advance buffer indexes.                               *)
          inc(XxBlockIndex, 4);
          inc(BinBuffIndex, 3);
          inc(BinBlockSize, 3);
          inc(XxBuffOffset, 4);

                   (* Until this XxBlock is completely decoded.             *)
        until (XxBlockIndex >= XxBlockSizeRec[TempBlockNum]);

                   (* Compare new Crc value with XxHeader Crc value.        *)
        if (CalcCrc <> CrcRec[TempBlockNum]) then
          begin
                   (* Erase the file being created if the Crc's don't match.*)
            EraseOutFile := true;
            ErrorMessage(18)
          end;

                   (* If this is not the last XxBlock in the XxBuffer then  *)
        if (TempBlockNum < TempIndex2) then

                   (* Set the XxBuffOffset index to the next XxBlock.       *)
          XxBuffOffset := BuffBlockPos[TempBlockNum + 1];

                   (* Advance the XxBlock-number indicator.                 *)
        inc(TempBlockNum, 1);

                   (* Until all XxBlocks in the XxBuffer are decoded.       *)
      until (BinBuffIndex >= BinBytesToWrite);

                   (* Set the binary-bytes to write to disk, after the last *)
                   (* XxBlock is processed.                                 *)
      if (TempBlockNum > XxBlockFound) then
        BinBytesToWrite := BinFilesize - BinFileIndex;

                   (* Write the BinBuffer bytes to disk.                    *)
      blockwrite(OutFile, BinBuffer, BinBytesToWrite);

                   (* Advance the BinFileIndex index.                       *)
      inc(BinFileIndex, BinBytesToWrite);

                   (* Until binary-file has been completely rebuilt.        *)
    until (BinFileIndex >= BinFilesize);
  end;

                   (*      MAIN PROGRAM BLOCK                               *)
BEGIN
  InitXx34;
  ProcessParams;
  OpenFiles;
  if Encoding then
    Encode34
  else
    Decode43;
  CloseFiles
END.


